/*
 * Copyright (c) 2012 Xilinx, Inc.  All rights reserved.
 *
 * Xilinx, Inc.
 * XILINX IS PROVIDING THIS DESIGN, CODE, OR INFORMATION "AS IS" AS A
 * COURTESY TO YOU.  BY PROVIDING THIS DESIGN, CODE, OR INFORMATION AS
 * ONE POSSIBLE   IMPLEMENTATION OF THIS FEATURE, APPLICATION OR
 * STANDARD, XILINX IS MAKING NO REPRESENTATION THAT THIS IMPLEMENTATION
 * IS FREE FROM ANY CLAIMS OF INFRINGEMENT, AND YOU ARE RESPONSIBLE
 * FOR OBTAINING ANY RIGHTS YOU MAY REQUIRE FOR YOUR IMPLEMENTATION.
 * XILINX EXPRESSLY DISCLAIMS ANY WARRANTY WHATSOEVER WITH RESPECT TO
 * THE ADEQUACY OF THE IMPLEMENTATION, INCLUDING BUT NOT LIMITED TO
 * ANY WARRANTIES OR REPRESENTATIONS THAT THIS IMPLEMENTATION IS FREE
 * FROM CLAIMS OF INFRINGEMENT, IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <pthread.h>
#include <signal.h>
#include "output.h"
#include "input.h"
#include "common_defines.h"

void InitalizeData();
void FillTestData(void);
int ReceivedCommand(void);
void RunCommand(void);
void RunPSOnlyTest(void);
void RunPLOnlyTest(void);
void RunPSandPLTest(void);
void ScaleCpuFreq(int freq);
pid_t spawn_Specload(int numCores);
int ReadCpuUtilization(void);
int read_fields (FILE *fp, unsigned long long int *fields);
char* exec1(const char* command);
void KillProcesses();
const char START_CMD[4] = "SOC";
const char END_CMD[4] = "EOC";
const char ERROR_RESPONSE[4] = "ERR";
const char START_OF_RESPONSE[4] = "SOR";
const char END_OF_RESPONSE[4] = "EOR";
pid_t pid_array[MAX_PROCESS];
pthread_t tid1,tid2;

unsigned int run_reponse_thread = TRUE;

void RunDemo(void);
int ext_ddr3_freq(void);


int main()
{

    printf("Zynq Power Demo\n");

	int handle;

	handle = open("/sys/devices/system/cpu/cpu0/cpufreq/scaling_governor", O_RDWR|O_SYNC );
	if (handle==-1)
	{
		error_print("ERROR: Unable to open /sys/devices/system/cpu/cpu0/cpufreq/scaling_governor");
		exit(1);
	}
	write(handle, "userspace", 10);
	close(handle);


//    while(1)
//    {
//    	ReadCpuUtilization();
//    //	ProcessResponseOutput(NULL);
//    }

    //test program - fills in the response buffer with test data
    //comment out on final version
    //FillTestData();
    //error_print("testing error response");
    InitalizeData();

    pthread_create(&tid1,NULL,(void*)ProcessCommandInput,NULL);
    //pthread_create(&tid2,NULL,(void*)ProcessResponseOutput,(void*)&run_reponse_thread);
    pthread_create(&tid2,NULL,(void*)ProcessResponseOutput,NULL);

    pthread_join(tid1,NULL);
    pthread_join(tid2,NULL);

    while(1)
    {
    	//nothing
    }

    return 0;
}

void RunCommand()
{
	switch(rxvd_command->test_num)
	{
		case PS_ONLY:
		{
			RunPSOnlyTest();
			break;
		}
		case PL_ONLY:
		{
			RunPLOnlyTest();
			break;
		}
		case PS_AND_PL:
		{
			RunPSandPLTest();
			break;
		}
		case DEMO_ONLY:
		{
			RunDemo();
		}
		case KILL_APP:
		{
			exit(127);
		}
	}
}

void FillTestData()
{
	int cur_rail;

	for(cur_rail=0;cur_rail < MAX_RAILS;cur_rail++)
	{
		pd_response.rail[cur_rail].voltage = 1.55;
		pd_response.rail[cur_rail].current = 0.34;
		pd_response.rail[cur_rail].power = 0.53;
	}

	pd_response.cpu_freq = 800;
	pd_response.ddr_freq = 667;
	pd_response.cpu0_util = 20;
	pd_response.cpu1_util = 50;
	pd_response.total_power = 2.38;
	pd_response.max_power = 3.5;
}

void InitalizeData()
{
	int cur_rail;
	int i;

	for(cur_rail=0;cur_rail < MAX_RAILS;cur_rail++)
	{
		pd_response.rail[cur_rail].voltage = 0.00;
		pd_response.rail[cur_rail].current = 0.00;
		pd_response.rail[cur_rail].power = 0.00;
	}

	pd_response.cpu_freq = 0;
	pd_response.ddr_freq = 0;
	pd_response.cpu0_util = 0;
	pd_response.cpu1_util = 0;
	pd_response.total_power = 0;
	pd_response.max_power = 0;

	//initialize the Array
	for(i=0;i<MAX_PROCESS;i++)
		pid_array[i]=0;
}

pid_t spawn_tasks(char *file, char *procName)
{
    pid_t pid = fork();
    if (pid == -1) {
        perror("fork");
        return -1;
    }

    /* when you call fork(), it creates two copies of your program:
     * a parent, and a child. you can tell them apart by the return
     * value from fork().  If fork() returns 0, this is is the child
     * process.  If fork() returns non-zero, we are the parent and the
     * return value is the PID of the child process. */
    if (pid == 0)
    {
        /* this is the child process.  now we can call one of the exec
         * family of functions to execute Spec.  when you call exec,
         * it replaces the currently running process (the child process)
         * with whatever we pass to exec.  So our child process will now
         * be running Spec load.exec() will never return except in an error
         * case, since it is now running the Spec code and not our code. */
    	printf("Executing %s with task name as %s\n",file,procName);
   		error_print("Spawaning tasks");
   		execl(file,procName,(char*)NULL);
   		perror(procName);
   		abort();
    }
    else
    {
        /* parent, return the child's PID back to main. */
        return pid;
    }
}

pid_t child_pid;

void RunPSOnlyTest()
{
//	int status;
//	int child_created = FALSE;
	int i;
	char file_arr[50],procName[50];

	for(i=0;i<sizeof(file_arr);i++)
	{
		file_arr[i]='\0';
		procName[i]='\0';
	}

	//create an array to hold list of PID's
	switch(rxvd_command->cpu_freq)
	{
		case FREQ_222:
			if(pd_response.cpu_freq != 222)
			{
				ScaleCpuFreq(222);
			}
			break;
		case FREQ_333:
			if(pd_response.cpu_freq != 333)
			{
				ScaleCpuFreq(333);
			}
			break;
		case FREQ_667:
			if(pd_response.cpu_freq != 666)
			{
				ScaleCpuFreq(666);
			}
			break;
		default:
			error_print("Selected frequency not supported");
			break;
	}

	switch(rxvd_command->cpu_load)
	{
		case CPU_LOAD_1:
		{
			KillProcesses();

			//sleep for a msec
			usleep(1000);
			break;
		}
		case CPU_LOAD_2:
		{
			KillProcesses();

			//run spec2000 on one cpu
			strcpy(file_arr,"/usr/games/freedoom");
			strcpy(procName,"freedoom");
			child_pid=spawn_tasks(file_arr,procName);
		    if (child_pid == -1) {
		    	error_print("failed to fork freedoom process\n");
		    }
		    printf("spawned freedoom with pid %d\n",child_pid);
		    //push the PID in to array
			for(i=0;i<MAX_PROCESS;i++)
			{
				if(pid_array[i] ==0)
				{
					pid_array[i]=child_pid;
					break;
				}
			}
		    break;
		}
		case CPU_LOAD_3:
		{
			KillProcesses();

			//run spec2000 on 2 cpu
			strcpy(file_arr,"dosbox /home/chandra/games/DAVE.EXE");
			strcpy(procName,"DAVE");
			child_pid=spawn_tasks(file_arr,procName);
		    if (child_pid == -1) {
		    	error_print("failed to fork DAVE process\n");
		    }
		    printf("spawned DAVE with pid %d\n",child_pid);
		    //push the PID in to array
			for(i=0;i<MAX_PROCESS;i++)
			{
				if(pid_array[i] ==0)
				{
					pid_array[i]=child_pid;
					break;
				}
			}
		    break;
		}
		case CPU_LOAD_4:
		{
			KillProcesses();
#if 0

			//We support wake up on GPIO
			int fp;
			error_print("Enabling wakup on GPIO support\n");
			fp = open("/sys/class/gpio/export", O_RDWR);
			if (!fp)
			{
				error_print("ERROR: Unable to open /sys/class/gpio/export");
				exit(1);
			}
			write(fp, "12", 3);
			close(fp);

			fp = open("/sys/class/gpio/export", O_RDWR);
			if (!fp)
			{
				error_print("ERROR: Unable to open /sys/class/gpio/export");
				exit(1);
			}
			write(fp, "14", 3);
			close(fp);

			fp = open("/sys/class/gpio/gpio12/edge", O_RDWR);
			if (!fp)
			{
				error_print("ERROR: Unable to open /sys/class/gpio/gpio12/edge");
				exit(1);
			}
			write(fp, "both", 5);
			close(fp);

			fp = open("/sys/class/gpio/gpio14/edge", O_RDWR);
			if (!fp)
			{
				error_print("ERROR: Unable to open /sys/class/gpio/gpio14/edge");
				exit(1);
			}
			write(fp, "both", 5);
			close(fp);

			fp = open("/sys/devices/amba.0/e000a000.ps7-gpio/power/wakeup", O_RDWR);
			if (!fp)
			{
				error_print("ERROR: Unable to open /sys/devices/amba.0/e000a000.ps7-gpio/power/wakeup");
				exit(1);
			}
			write(fp, "enabled", 8);
			close(fp);

			error_print("Entering suspend Mode\n");

			fp = open("/sys/power/state", O_RDWR);
			if (!fp)
			{
				error_print("ERROR: Unable to open /sys/power/state");
				exit(1);
			}
			write(fp, "mem", 4);

			close(fp);

/*			strcpy(file_arr,"/mnt/standby.sh");
			strcpy(procName,"standby");
			child_pid=spawn_tasks(file_arr,procName);
		    if (child_pid == -1) {
		    	error_print("failed to fork standby process\n");
		    }
*/
		    /*printf("spawned spec-2-core with pid %d\n",child_pid);
		    //push the PID in to array
			for(i=0;i<MAX_PROCESS;i++)
			{
				if(pid_array[i] ==0)
				{
					pid_array[i]=child_pid;
					break;
				}
			}
		    break;
		    */
		    break;
#endif
		}
	}

}


void RunPLOnlyTest()
{

}

void RunPSandPLTest()
{

}

void ScaleCpuFreq(int freq)
{
	int cpu0_freq_handle;

	cpu0_freq_handle = open("/sys/devices/system/cpu/cpu0/cpufreq/scaling_setspeed", O_RDWR|O_SYNC );
	if (cpu0_freq_handle==-1)
	{
		error_print("ERROR: Unable to open /sys/devices/system/cpu/cpu0/cpufreq/scaling_setspeed");
		exit(1);
	}
	if (freq == 666)
	{
		write(cpu0_freq_handle, "666667", 7);
	}
	else if(freq == 333)
	{
		write(cpu0_freq_handle, "333334", 7);
	}
	else if(freq == 222)
	{
		write(cpu0_freq_handle, "222223", 7);
	}

	close(cpu0_freq_handle);
}

void ReadAdditionalData(void)
{
	char cpu_freq_string[20];
	int cpu_freq;
	int cpu_freq_handle;

	cpu_freq_handle = open("/sys/devices/system/cpu/cpu0/cpufreq/cpuinfo_cur_freq", O_RDONLY);
	if (cpu_freq_handle == -1)
	{
		error_print("ERROR: Unable to open /sys/devices/system/cpu/cpu0/cpufreq/cpuinfo_cur_freq");
	}

	read(cpu_freq_handle, cpu_freq_string, 7);
	cpu_freq = atoi(cpu_freq_string);
	pd_response.cpu_freq = cpu_freq/1000;

	//Read DDR frequency
	pd_response.ddr_freq = ext_ddr3_freq();
	//pd_response.ddr_freq = 667;

	close(cpu_freq_handle);

}

int read_fields (FILE *fp, unsigned long long int *fields)
{
  int retval;
  char buffer[BUF_MAX];


  if (!fgets (buffer, BUF_MAX, fp))
  { perror ("Error"); }
  /* line starts with c and a string. This is to handle cpu, cpu[0-9]+ */
  retval = sscanf (buffer, "c%*s %Lu %Lu %Lu %Lu %Lu %Lu %Lu %Lu %Lu %Lu",
                            &fields[0],
                            &fields[1],
                            &fields[2],
                            &fields[3],
                            &fields[4],
                            &fields[5],
                            &fields[6],
                            &fields[7],
                            &fields[8],
                            &fields[9]);
  if (retval == 0)
  { return -1; }
  if (retval < 4) /* Atleast 4 fields is to be read */
  {
    fprintf (stderr, "Error reading /proc/stat cpu field\n");
    return 0;
  }
  return 1;
}

//int ReadCpuUtilization(void)
//{
//	  FILE *fp;
//	  unsigned long long int fields[10], total_tick[MAX_CPU], total_tick_old[MAX_CPU], idle[MAX_CPU], idle_old[MAX_CPU], del_total_tick[MAX_CPU], del_idle[MAX_CPU];
//	  int update_cycle = 0, i, cpus = 0, count;
//	  double percent_usage;
//
//	  fp = fopen ("/proc/stat", "r");
//	  if (fp == NULL)
//	  {
//	    perror ("Error");
//	  }
//
//	  while (read_fields (fp, fields) != -1)
//	  {
//	    for (i=0, total_tick[cpus] = 0; i<10; i++)
//	    {
//	    	total_tick[cpus] += fields[i];
//	    }
//	    idle[cpus] = fields[3]; /* idle ticks index */
//	    cpus++;
//	  }
//
//		fseek (fp, 0, SEEK_SET);
//		fflush (fp);
//		printf ("[Update cycle %d]\n", update_cycle);
//		for (count = 0; count < cpus; count++)
//		{
//			total_tick_old[count] = total_tick[count];
//			idle_old[count] = idle[count];
//
//			if (!read_fields (fp, fields))
//			{
//				error_print("[read_fields Failed]\n");
//				return 0;
//			}
//
//			for (i=0, total_tick[count] = 0; i<10; i++)
//			{
//				total_tick[count] += fields[i];
//			}
//			idle[count] = fields[3];
//			del_total_tick[count] = total_tick[count] - total_tick_old[count];
//			del_idle[count] = idle[count] - idle_old[count];
//
//			percent_usage = ((del_total_tick[count] - del_idle[count]) / (double) del_total_tick[count]) * 200;
//
//			if(count == 0)
//				pd_response.cpu0_util=percent_usage;
//			else
//				pd_response.cpu1_util=percent_usage;
//
//			if (count == 0)
//			{
//				printf ("Total CPU Usage: %3.2lf%%\n", percent_usage);
//			}
//			else
//			{
//				printf ("\tCPU%d Usage: %3.2lf%%\n", count - 1, percent_usage);
//			}
//		}
//		update_cycle++;
//			/* Ctrl + C quit, therefore this will not be reached. We rely on the kernel to close this file */
//			fclose (fp);
//	//i have to also read ddr frequency and cpu utilization.
//		pd_response.ddr_freq = 667;
//		return 0;
//}


void KillProcesses(void)
{
	int i;
	char killString[40];
	char pid_string[10];

	//check to see if any pid's are running. if yes then kill them first
	for(i=0;i<MAX_PROCESS;i++)
	{
	   /* kill will send the specified signal to the specified process.
		 * in this case, we send a TERM signal to VLC, requesting that it
		 * terminate.  If that doesn't work, we send a KILL signal.
		 * If that doesn't work, we give up.
		 **/
		if(pid_array[i] != 0)
		{
			printf("Killing the child processes spawned earlier has PID %d\n",pid_array[i]);
			kill(pid_array[i],SIGQUIT);
			kill(pid_array[i],SIGTERM);
			usleep(1000);
			kill(pid_array[i],SIGKILL);
			strcpy(killString, "killall ./crafty_base.gcc48-default");;
			system(killString);

			strcpy(killString, "killall ./crafty_base_c.gcc48-default");;
			system(killString);

			printf(killString);
//			if (kill(pid_array[i],SIGKILL) < 0)
//			{
//				printf("kill with SIGTERM");
//				if(kill(pid_array[i],SIGKILL) < 0)
//				{
//					printf("kill with SIGKILL");
//				}
//			}
		}
		pid_array[i] = 0;
	}
}


int ext_ddr3_freq()
{
	FILE *fp;
	long long  ret_val=400;
	char buf[80];
/*
	//fp=fopen("/sys/kernel/debug/clk/ps_clk/ddrpll_int/ddrpll/ddr3x_div/ddr3/clk_rate","r");

	if(fp==NULL)
	{
		error_print("error opening /sys/kernel/debug/clk/ps_clk/ddrpll_int/ddrpll/ddr3x_div/ddr3/clk_rate");
		return 533;
	}

	fgets(buf,80,fp);
	printf("buf=%s\n",buf);
	ret_val=atoi(buf);
	//printf("DDR value=%lld\n",ret_val);
*/
	return ret_val;
}

void RunDemo(void)
{
	int i;
	char file_arr[50],procName[50];

	while(ReceivedCommand() != TRUE)
	{
		//run spec-2000  on 2 cores
		//at 666 MHz
		ScaleCpuFreq(666);
		//buzz_print("Running spec2000 on 2 cores @ 666 MHz");
		//system("./mnt/spec2000-2-core.sh");
		//run spec2000 on 2 cpu
		strcpy(file_arr,"/mnt/spec2000-1-core.sh");
		strcpy(procName,"spec2000_2_0_core");
		child_pid=spawn_tasks(file_arr,procName);
		if (child_pid == -1) {
			error_print("failed to fork spec-2-core process\n");
		}
		printf("spawned spec-2-core with pid %d\n",child_pid);
		//push the PID in to array
		for(i=0;i<MAX_PROCESS;i++)
		{
			if(pid_array[i] ==0)
			{
				pid_array[i]=child_pid + 2;
				break;
			}
		}
		strcpy(file_arr,"/mnt/spec2000-2-core.sh");
		strcpy(procName,"spec2000_2_1_core");
		child_pid=spawn_tasks(file_arr,procName);
		if (child_pid == -1) {
			error_print("failed to fork spec-2-core process\n");
		}
		printf("spawned spec-2-core with pid %d\n",child_pid);
		//push the PID in to array
		for(i=0;i<MAX_PROCESS;i++)
		{
			if(pid_array[i] ==0)
			{
				pid_array[i]=child_pid + 2;
				break;
			}
		}

		sleep(60);
		//let spec complete - i think spec will take 5 minutes?

		//run cpu idle at 222 MHz
		//wait for 3 minutes
		KillProcesses();
		ScaleCpuFreq(222);
		buzz_print("CPU idle task @ 222 MHz");
		sleep(60);
	}

	//out of the loop
	//inform upender we are out
	printf("DefaultDemoEnd");
}

